/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: VMwareServerPool.java 603 2011-11-02 17:59:37Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.vmware;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.FutureTask;

import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.AbstractStoragePool;
import org.ow2.sirocco.vmm.agent.domain.ManagedResource;
import org.ow2.sirocco.vmm.agent.domain.ServerPool;
import org.ow2.sirocco.vmm.agent.driver.util.RemoteExec;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VirtualMachineImageMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineImageStoreMXBean;

import com.vmware.vim.ArrayOfManagedObjectReference;
import com.vmware.vim.ArrayOfString;
import com.vmware.vim.DynamicProperty;
import com.vmware.vim.ManagedObjectReference;

/**
 * VMware driver ServerPool MXBean implementation
 */
class VMwareServerPool extends ServerPool {
    static protected Logger logger = Logger.getLogger(VMwareServerPool.class);

    static private int KEEP_ALIVE_PERIOD_MS = 2 * 60 * 1000;

    private static final int THREADPOOL_SIZE = 20;

    private static final int CONNECTION_DEFAULT_POOL_SIZE = 4;

    private ExecutorService executorService = Executors.newFixedThreadPool(VMwareServerPool.THREADPOOL_SIZE);

    private ConnectionPool connectionPool;

    private HashMap<String, VMwareHost> hostRefs = new HashMap<String, VMwareHost>();

    private String datacenter, vmFolderName, datastoreName, networkName, networkDHCPName, networkStaticIPName, resourcePool;

    private ManagedObjectReference datacenterRef, vmFolderRef, datastoreRef, vmTemplateFolderRef, hostFolderRef,
        resourcePoolRef;;

    private VMwareVMTemplateDataStore imageStore;

    private RemoteExec.SshAuthInfo defaultAuthInfo;

    private VMwareEventCollector eventCollector;

    public VMwareServerPool(final String name, final ObjectName objectName, final Map<String, String> attributes)
        throws IllegalArgumentException, VMMException {
        super(name, objectName, attributes);
        String hostName = attributes.get("vCenterHostName");
        if (hostName == null) {
            throw new IllegalArgumentException("Missing parameter: vCenterHostName");
        }
        this.datacenter = attributes.get("datacenter");
        if (this.datacenter == null) {
            throw new IllegalArgumentException("Missing parameter: datacenter");
        }
        this.vmFolderName = attributes.get("vmFolderPath");
        String vmTemplateFolderName = attributes.get("vmTemplateFolderPath");

        String userName = attributes.get("user");
        if (userName == null) {
            throw new IllegalArgumentException("Missing parameter: userName");
        }
        String password = attributes.get("password");
        if (password == null) {
            throw new IllegalArgumentException("Missing parameter: password");
        }
        this.datastoreName = attributes.get("datastore");
        this.networkName = attributes.get("network");
        this.networkDHCPName = attributes.get("networkDHCP");
        this.networkStaticIPName = attributes.get("networkStaticIP");

        this.resourcePool = attributes.get("resourcePool");

        String sshPassword = attributes.get("sshDefaultRootPassword");
        String sshPrivateKeyFile = attributes.get("sshDefaultPrivateKeyFile");
        try {
            this.defaultAuthInfo = new RemoteExec.SshAuthInfo(sshPassword, sshPrivateKeyFile);
        } catch (IllegalArgumentException ex) {
            // XXX
        }

        VMwareServiceConnection connection = null;
        try {
            this.connectionPool = new ConnectionPool(this.executorService, VMwareServerPool.CONNECTION_DEFAULT_POOL_SIZE,
                hostName, userName, password);
            connection = this.connectionPool.getConnection();

            VMwareServerPool.logger.debug("Finding datacenter " + this.datacenter);
            this.datacenterRef = connection.getService().findByInventoryPath(connection.getServiceContent().getSearchIndex(),
                this.datacenter);
            if (this.datacenterRef == null) {
                String msg = "VMware Driver initialization error: datacenter " + this.datacenter + " not found ";
                VMwareServerPool.logger.error(msg);
                throw new VMMException(msg);
            }

            if (this.vmFolderName == null) {
                this.vmFolderRef = (ManagedObjectReference) connection.getDynamicProperties(this.datacenterRef, "vmFolder")[0]
                    .getVal();
            } else {
                this.vmFolderRef = this.getVmFolder(connection, this.vmFolderName);
            }

            // Find the virtual machine template folder
            if (vmTemplateFolderName == null) {
                this.vmTemplateFolderRef = this.vmFolderRef;
            } else {
                this.vmTemplateFolderRef = this.getVmFolder(connection, vmTemplateFolderName);
            }

            // Find the host folder for this folder.
            this.hostFolderRef = (ManagedObjectReference) connection.getDynamicProperties(this.datacenterRef, "hostFolder")[0]
                .getVal();
            if (this.hostFolderRef == null) {
                String msg = "VMware Driver initialization error: host folder not found for datacenter " + this.datacenter;
                VMwareServerPool.logger.error(msg);
                throw new VMMException(msg);
            }

            // find the datastore
            if (this.datastoreName != null) {
                ArrayOfManagedObjectReference dataStoreRefs = (ArrayOfManagedObjectReference) connection.getDynamicProperties(
                    this.datacenterRef, "datastore")[0].getVal();
                if (dataStoreRefs != null) {
                    for (int i = 0; i < dataStoreRefs.getManagedObjectReference().length; i++) {
                        String dsName = (String) connection.getDynamicProperties(dataStoreRefs.getManagedObjectReference(i),
                            "info.name")[0].getVal();
                        if (dsName.equals(this.datastoreName)) {
                            this.datastoreRef = dataStoreRefs.getManagedObjectReference(i);
                            break;
                        }
                    }
                }
            }

            // find ResourcePool ref
            if (this.resourcePool != null) {
                this.resourcePoolRef = connection.getService().findByInventoryPath(
                    connection.getServiceContent().getSearchIndex(), this.resourcePool);
                if (this.resourcePoolRef == null) {
                    VMwareServerPool.logger.warn("Cannot find ResourcePool " + this.resourcePool);
                }
            }

        } catch (Exception ex) {
            String msg = "VMware Driver initialization error: " + ex.getMessage();
            VMwareServerPool.logger.error(msg, ex);
            throw new VMMException(msg);
        } finally {
            if (connection != null) {
                connection.release();
            }
        }

        this.eventCollector = new VMwareEventCollector(this, new VMwareServiceConnection(hostName, userName, password),
            this.datacenter);
        this.eventCollector.start();

    }

    private ManagedObjectReference getVmFolder(final VMwareServiceConnection connection, final String folderName)
        throws Exception {
        ManagedObjectReference folderRef = connection.getService().findByInventoryPath(
            connection.getServiceContent().getSearchIndex(), folderName);
        if (folderRef == null) {
            throw new Exception("VMware Driver initialization error: virtual machine folder " + folderName + " not found");
        }
        if (!folderRef.getType().equals("Folder")) {
            throw new Exception(folderName + " is not a folder");
        }
        ArrayOfString childTypes = (ArrayOfString) connection.getDynamicProperties(folderRef, "childType")[0].getVal();
        for (String childType : childTypes.getString()) {
            if (childType.equals("VirtualMachine")) {
                return folderRef;
            }
        }
        throw new Exception("Folder " + folderName + " is a not a VirtualMachine folder");
    }

    @Override
    public void postCreateHook() {
        super.postCreateHook();
        for (StoragePoolMXBean storagePool : this.getStoragePools()) {
            try {
                this.storageCapacityMB += storagePool.getCapacityMB();
            } catch (VMMException e) {
                continue;
            }
        }
    }

    @Override
    public void release() {
        this.executorService.shutdown();
        this.eventCollector.stop();
        this.connectionPool.stop();
        this.connectionPool = null;
    }

    @Override
    public String getHypervisor() {
        return "ESX";
    }

    public ExecutorService getExecutorService() {
        return this.executorService;
    }

    public ConnectionPool getConnectionPool() {
        return this.connectionPool;
    }

    @Override
    public VirtualMachineImageStoreMXBean getVMImageStore() {
        return this.imageStore;
    }

    @Override
    protected AbstractStoragePool newStoragePool(final String name, final ObjectName objectName,
        final java.util.Map<String, String> attributes) throws Exception {
        String dataStoreName = attributes.get("dataStoreName");
        String filePath = attributes.get("filePath");
        if (dataStoreName == null) {
            throw new Exception("Missing attribute: dataStoreName");
        }
        if (filePath == null) {
            throw new Exception("Missing attribute: filePath");
        }
        int flags = VMwareStoragePool.TEMPLATE_DISK | VMwareStoragePool.ISO_IMAGE | VMwareStoragePool.VM_DISK;
        return new VMwareStoragePool(name, objectName, this, dataStoreName, filePath, flags, attributes);
    }

    public String getMountPoint() {
        return this.datacenter;
    }

    VMwareHost getHostByName(final String name) {
        return this.hostRefs.get(name);
    }

    @Override
    public HostMXBean newHost(final String hostName, final Map<String, String> props) throws VMMException {
        ManagedObjectReference mor = null;
        VMwareServiceConnection connection = this.connectionPool.getConnection();
        try {
            mor = connection.getDecendentMoRef(this.getHostFolderRef(), "HostSystem", hostName);
        } catch (Exception ex) {
            VMwareServerPool.logger.error("Unable to find HostSystem managed object " + hostName + " in VMware VC inventory",
                ex);
            return null;
        } finally {
            connection.release();
        }

        if (mor == null) {
            VMwareServerPool.logger.error("Host " + hostName + " not found");
            return null;
        }

        ObjectName mbeanObjectName = MBeanObjectNamer.makeHostName(this.getPath() + "/" + hostName, hostName);
        VMwareHost host;
        try {
            host = new VMwareHost(this, mor, mbeanObjectName, hostName, props);
        } catch (Exception ex) {
            throw new VMMException("Failed to add host " + hostName + ex.getMessage());
        }
        host.getResidentVMs();

        try {
            AgentCommon.getMBeanServer().registerMBean(host, mbeanObjectName);
        } catch (Exception ex) {
            VMwareServerPool.logger.error("Unable to register MBean " + mbeanObjectName, ex);
            return null;
        }
        this.hostRefs.put(hostName, host);
        // logger.info("Added host "+hostName+" "+host.getObjectName());
        this.addHost(host);
        try {
            this.connectionPool.ensureCapacity(2 + this.hostRefs.size() * 3); // XXX
        } catch (Exception ex) {
            VMwareServerPool.logger.error("Cannot increase connection pool", ex);
        }
        return host;
    }

    public ManagedObjectReference getDatacenterRef() {
        return this.datacenterRef;
    }

    public ManagedObjectReference getVmFolderRef() {
        return this.vmFolderRef;
    }

    public ManagedObjectReference getHostFolderRef() {
        return this.hostFolderRef;
    }

    public ManagedObjectReference getVmTemplateFolderRef() {
        return this.vmTemplateFolderRef;
    }

    public ManagedObjectReference getDatastoreRef() {
        return this.datastoreRef;
    }

    public ManagedObjectReference getResourcePoolRef() {
        return this.resourcePoolRef;
    }

    public String getDatastoreName() {
        return this.datastoreName;
    }

    public String getNetworkName() {
        return this.networkName;
    }

    public String getNetworkDHCPName() {
        return this.networkDHCPName;
    }

    public String getNetworkStaticIPName() {
        return this.networkStaticIPName;
    }

    static class ConnectionPool {
        private List<VMwareServiceConnection> connections = new ArrayList<VMwareServiceConnection>();

        private String hostName, userName, password;

        private volatile boolean stopped = false;

        private Callable<Void> keepAliveCallable = new Callable<Void>() {
            public Void call() {
                while (!ConnectionPool.this.stopped) {
                    try {
                        synchronized (ConnectionPool.this) {
                            for (VMwareServiceConnection connection : ConnectionPool.this.connections) {
                                if (connection.lease()) {
                                    try {
                                        Calendar time = connection.getService().currentTime(connection.getServiceInstanceRef());
                                    } finally {
                                        connection.release();
                                    }
                                }
                            }
                        }
                    } catch (Exception ex) {
                    }
                    if (!Thread.currentThread().isInterrupted() && !ConnectionPool.this.stopped) {
                        try {
                            Thread.sleep(VMwareServerPool.KEEP_ALIVE_PERIOD_MS);
                        } catch (InterruptedException e) {
                            VMwareServerPool.logger.debug("KeepAlive task interrupted");
                            break;
                        }
                    } else {
                        break;
                    }
                }
                return null;
            }
        };

        private FutureTask<Void> keepAliveTask = new FutureTask<Void>(this.keepAliveCallable);

        ConnectionPool(final ExecutorService executorService, final int initialCapacity, final String hostName,
            final String userName, final String password) throws Exception {
            this.userName = userName;
            this.hostName = hostName;
            this.password = password;
            this.ensureCapacity(initialCapacity);
            // executorService.execute(new Runnable() {
            // public void run() {
            // while (!ConnectionPool.this.stopped) {
            // try {
            // synchronized (ConnectionPool.this) {
            // for (VMwareServiceConnection connection :
            // ConnectionPool.this.connections) {
            // if (connection.lease()) {
            // try {
            // Calendar time = connection.getService().currentTime(
            // connection.getServiceInstanceRef());
            // } finally {
            // connection.release();
            // }
            // }
            // }
            // }
            // Thread.sleep(VMwareServerPool.KEEP_ALIVE_PERIOD_MS);
            // } catch (Exception ex) {
            // VMwareServerPool.logger.error("error", ex);
            // }
            // }
            // }
            // });
            executorService.submit(this.keepAliveTask);
        }

        synchronized void ensureCapacity(final int minCapacity) throws Exception {
            if (this.connections.size() < minCapacity) {
                for (int i = minCapacity - this.connections.size(); i > 0; i--) {
                    VMwareServiceConnection connection = new VMwareServiceConnection(this.hostName, this.userName,
                        this.password);
                    connection.connect();
                    this.connections.add(connection);
                }
            }
        }

        private synchronized VMwareServiceConnection getFirstFreeConnection() {
            for (VMwareServiceConnection connection : this.connections) {
                if (connection.lease()) {
                    return connection;
                }
            }
            return null;
        }

        synchronized void stop() {
            this.stopped = true;
            this.keepAliveTask.cancel(true);
            for (VMwareServiceConnection connection : this.connections) {
                try {
                    connection.close();
                } catch (Exception ex) {
                }
            }
            this.connections.clear();
        }

        synchronized VMwareServiceConnection getConnection() {
            VMwareServiceConnection connection = this.getFirstFreeConnection();
            if (connection == null) {
                try {
                    this.ensureCapacity(this.connections.size() + 5);
                } catch (Exception ex) {
                    // ignore
                }
                connection = this.getFirstFreeConnection();
            }
            return connection;
        }
    }

    class VMwareVMTemplateDataStore extends ManagedResource implements VirtualMachineImageStoreMXBean {
        private String name;

        private ArrayList<VirtualMachineImageMXBean> imageList = new ArrayList<VirtualMachineImageMXBean>();

        private ManagedObjectReference datastoreRef;

        private boolean needSync = true;

        public VMwareVMTemplateDataStore(final ObjectName objectName, final String name,
            final ManagedObjectReference datastoreRef, final Map<String, String> attributes) {
            super(objectName, null);
            this.name = name;
            this.datastoreRef = datastoreRef;
            this.listVMImageTemplates();
            Set<String> templateNames = new HashSet<String>();
            for (String key : attributes.keySet()) {
                if (key.startsWith("template")) {
                    templateNames.add(key.substring(0, key.indexOf('.')));
                }
            }
            for (String template : templateNames) {
                String isoFileName = attributes.get(template + ".isoFileName");
                String description = attributes.get(template + ".description");
                String uuid = attributes.get(template + ".uuid");
                String makeFloppyShellScript = attributes.get(template + ".makeFloppyShellScript");
                String guestId = attributes.get(template + ".guestId");

                ObjectName on = MBeanObjectNamer.makeVMImageName(uuid);
                VMwareVirtualMachineISOTemplate image = new VMwareVirtualMachineISOTemplate(on, uuid, uuid);
                image.setMakeFloppyShellScript(makeFloppyShellScript);
                image.setGuestId(guestId);
                image.setIsoFileName(isoFileName);
                VMwareServerPool.logger.debug("New VMware template: " + image);
                try {
                    AgentCommon.getMBeanServer().registerMBean(image, on);
                } catch (Exception ex) {
                    VMwareServerPool.logger.error("Failed to register image", ex);
                    continue;
                }
                this.imageList.add(image);

            }
        }

        public long getCapacityMB() {
            VMwareServiceConnection connection = VMwareServerPool.this.connectionPool.getConnection();
            try {
                long capacityInBytes = (Long) connection.getDynamicProperties(this.datastoreRef, "summary.capacity")[0]
                    .getVal();
                return capacityInBytes / (1024 * 1024);
            } catch (Exception ex) {
                VMwareServerPool.logger.error("Failed to get datastore capacity", ex);
                return -1;
            } finally {
                connection.release();
            }
        }

        public long getFreeSpaceMB() {
            VMwareServiceConnection connection = VMwareServerPool.this.connectionPool.getConnection();
            try {
                long freeSpaceInBytes = (Long) connection.getDynamicProperties(this.datastoreRef, "summary.freeSpace")[0]
                    .getVal();
                return freeSpaceInBytes / (1024 * 1024);
            } catch (Exception ex) {
                VMwareServerPool.logger.error("Failed to get datastore free space", ex);
                return -1;
            } finally {
                connection.release();
            }
        }

        public String getName() {
            return this.name;
        }

        public synchronized void sync() {
            this.needSync = true;
        }

        public synchronized List<VirtualMachineImageMXBean> listVMImageTemplates() {
            VMwareServerPool.logger.debug("listVMImages");
            if (this.needSync) {
                VMwareServiceConnection connection = VMwareServerPool.this.connectionPool.getConnection();
                try {
                    ArrayList<ManagedObjectReference> vms = connection.getDecendentMoRefs(
                        VMwareServerPool.this.getVmTemplateFolderRef(), "VirtualMachine");

                    for (ManagedObjectReference vmMor : vms) {
                        boolean isTemplate = (Boolean) connection.getDynamicProperties(vmMor, "summary.config.template")[0]
                            .getVal();
                        if (isTemplate) {
                            DynamicProperty[] properties = connection.getDynamicProperties(vmMor, "config.name");
                            if (properties == null || properties.length < 1) {
                                continue;
                            }
                            String vmName = (String) properties[0].getVal();
                            String uuid = (String) connection.getDynamicProperties(vmMor, "config.uuid")[0].getVal();

                            boolean found = false;
                            for (VirtualMachineImageMXBean image : this.imageList) {
                                if (image.getUUID().equals(vmName)) {
                                    found = true;
                                }
                            }
                            if (!found) {
                                ObjectName on = MBeanObjectNamer.makeVMImageName(uuid);
                                VMwareVirtualMachineTemplate image = new VMwareVirtualMachineTemplate(on, vmName, uuid);
                                VMwareServerPool.logger.debug("New VMware template: " + image);
                                AgentCommon.getMBeanServer().registerMBean(image, on);
                                this.imageList.add(image);
                            }
                        }

                    }

                } catch (Exception ex) {
                    VMwareServerPool.logger.error("Failed to list VM templates", ex);
                } finally {
                    connection.release();
                }
                this.needSync = false;
            }
            return this.imageList;
        }

        public synchronized VirtualMachineImageMXBean lookUpByUUID(final String uuid) {
            VMwareServerPool.logger.debug("lookUpByUUID " + uuid);
            for (VirtualMachineImageMXBean image : this.listVMImageTemplates()) {
                if (image.getUUID().equals(uuid)) {
                    return image;
                }
            }
            return null;
        }

        public synchronized void removeVMImageTemplate(final VirtualMachineImageMXBean vmImage) throws VMMException {
            String vmImageID = vmImage.getUUID();
            VMwareServiceConnection connection = VMwareServerPool.this.connectionPool.getConnection();
            try {
                ManagedObjectReference vmMOR = connection.getDecendentMoRef(null, "VirtualMachine", vmImageID);
                // asynchronous op, forget about result ?
                connection.getService().destroy_Task(vmMOR);
            } catch (Exception ex) {
                VMwareServerPool.logger.error("Failed to retrieve VM reference " + vmImageID, ex);
            } finally {
                connection.release();
            }
        }

    }

    static class VMwareVirtualMachineTemplate extends ManagedResource implements VirtualMachineImageMXBean {
        private String name, uuid;

        public VMwareVirtualMachineTemplate(final ObjectName objectName, final String name, final String uuid) {
            super(objectName, null);
            this.name = name;
            this.uuid = uuid;
        }

        public String getName() {
            return this.name;
        }

        // TODO we ignore the UUID and use the name as a unique
        public String getUUID() {
            return this.name;
        }

        public Map<String, String> getMetaData() {
            return null;
        }

        public String getDescription() {
            return "";
        }

        @Override
        public String toString() {
            return "VMwareTemplate(name=" + this.getName() + ",uuid=" + this.getUUID() + ")";
        }
    }

    static class VMwareVirtualMachineISOTemplate extends VMwareVirtualMachineTemplate implements VirtualMachineImageMXBean {
        private String isoFileName, makeFloppyShellScript, guestId;

        public VMwareVirtualMachineISOTemplate(final ObjectName objectName, final String name, final String uuid) {
            super(objectName, name, uuid);
        }

        public String getIsoFileName() {
            return this.isoFileName;
        }

        public void setIsoFileName(final String isoFileName) {
            this.isoFileName = isoFileName;
        }

        public String getMakeFloppyShellScript() {
            return this.makeFloppyShellScript;
        }

        public void setMakeFloppyShellScript(final String makeFloppyShellScript) {
            this.makeFloppyShellScript = makeFloppyShellScript;
        }

        public String getGuestId() {
            return this.guestId;
        }

        public void setGuestId(final String guestId) {
            this.guestId = guestId;
        }

        @Override
        public String toString() {
            return "VMwareISOTemplate(uuid=" + this.getUUID() + ",iso=" + this.getIsoFileName() + ")";
        }

    }

    @Override
    public void deleteImageStore() {
        try {
            for (VirtualMachineImageMXBean image : this.imageStore.imageList) {
                AgentCommon.getMBeanServer().unregisterMBean(image.getObjectName());
            }
            this.imageStore.imageList.clear();
        } catch (Throwable e) {
            ServerPool.logger.error(e);
        }
        try {
            AgentCommon.getMBeanServer().unregisterMBean(this.imageStore.getObjectName());
        } catch (Exception e) {
            ServerPool.logger.error(e);
        }
        this.imageStore = null;
    }
}
