/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: VMwareVirtualMachine.java 828 2012-01-11 22:41:32Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.vmware;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.AbstractVirtualMachine;
import org.ow2.sirocco.vmm.api.BadVMPowerStateException;
import org.ow2.sirocco.vmm.api.DiskOperation;
import org.ow2.sirocco.vmm.api.GuestInfo;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.IllegalOperationException;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.ResourceUsage;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMCustomizationSpec;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VNIC;
import org.ow2.sirocco.vmm.api.VirtualDisk;
import org.ow2.sirocco.vmm.api.VirtualDiskSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineImageMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean;
import org.ow2.sirocco.vmm.api.Volume;

import com.vmware.vim.ArrayOfInt;
import com.vmware.vim.ArrayOfVirtualDevice;
import com.vmware.vim.DynamicProperty;
import com.vmware.vim.InvalidPowerState;
import com.vmware.vim.ManagedObjectReference;
import com.vmware.vim.ObjectContent;
import com.vmware.vim.TaskInfo;
import com.vmware.vim.VirtualCdrom;
import com.vmware.vim.VirtualCdromIsoBackingInfo;
import com.vmware.vim.VirtualDevice;
import com.vmware.vim.VirtualDeviceConfigSpec;
import com.vmware.vim.VirtualDeviceConfigSpecOperation;
import com.vmware.vim.VirtualDeviceConnectInfo;
import com.vmware.vim.VirtualDeviceFileBackingInfo;
import com.vmware.vim.VirtualDiskFlatVer2BackingInfo;
import com.vmware.vim.VirtualEthernetCard;
import com.vmware.vim.VirtualHardware;
import com.vmware.vim.VirtualMachineCloneSpec;
import com.vmware.vim.VirtualMachineConfigSpec;
import com.vmware.vim.VirtualMachineMovePriority;
import com.vmware.vim.VirtualMachinePowerState;
import com.vmware.vim.VirtualMachineRelocateSpec;
import com.vmware.vim.VirtualSCSIController;

/**
 * VMware driver VirtualMachine MXBean implementation
 */
class VMwareVirtualMachine extends AbstractVirtualMachine implements VirtualMachineMXBean {
    static Logger logger = Logger.getLogger(VMwareVirtualMachine.class);

    private VMwareServerPool pool;

    private VMwareHost host;

    private ManagedObjectReference ref;

    private String nameLabel;

    private String uuid;

    private String macAddress;

    private String ipAddress;

    private int numVcpus = -1;

    private long memorySizeMB = -1;

    HashMap<String, BigInteger> metricValues = null;

    private List<VirtualDisk> disks;

    private List<org.ow2.sirocco.vmm.api.VirtualCdrom> cdroms;

    public VMwareVirtualMachine(final ObjectName objectName, final VMwareServerPool pool, final ManagedObjectReference mor,
        final VMwareHost host, final String nameLabel) {
        super(objectName);
        this.pool = pool;
        this.ref = mor;
        this.host = host;
        this.uuid = null;
        this.nameLabel = nameLabel;
    }

    public ResourceUsage getResourceUsage() {
        ResourceUsage ru = new ResourceUsage();
        // XX approximate timestamp
        ru.setSamplingTime(new Date(System.currentTimeMillis()));
        ru.setCpuLoad(this.getCPULoad());
        ru.setMemoryUsedKBytes(this.getMemoryUsedMB() * 1024);
        // TODO net and memory perf
        return ru;
    }

    public void updateMetrics(final HashMap<String, BigInteger> m) {
        this.metricValues = m;
    }

    public boolean canLiveMigrateToHost(final HostMXBean targetHost) {
        // TODO
        return false;
    }

    public ManagedObjectReference getRef() {
        return this.ref;
    }

    public void migrate(HostMXBean targetHost, final boolean live) throws IllegalOperationException, VMMException {
        String targetHostName = targetHost.getHostName();
        VMwareVirtualMachine.logger.info("Attempting live migration of VM " + this.nameLabel + " to host " + targetHostName);

        boolean foundHostInServerPool = false;

        for (HostMXBean h : this.host.getServerPool().getManagedHosts()) {
            if (h.getHostName().equals(targetHost.getHostName())) {
                foundHostInServerPool = true;
                targetHost = h;
                break;
            }
        }

        if (!foundHostInServerPool) {
            throw new IllegalOperationException("Source and target hosts belong to different server pools");
        }

        VMwareHost targetVMwareHost = (VMwareHost) targetHost;

        VMwareVirtualMachine.logger.info("VM " + this.nameLabel + " live migration to host " + targetHostName + "...");
        this.emitNotification(NotificationTypes.VM_MIGRATION_START, targetHostName, null);

        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            ManagedObjectReference taskmor = connection.getService().migrateVM_Task(this.ref, null, targetVMwareHost.getMOR(),
                VirtualMachineMovePriority.defaultPriority, null);
            boolean result = this.getTaskInfo(connection, taskmor);
            if (result) {
                VMwareVirtualMachine.logger.debug("Virtual Machine " + this.nameLabel + " migrated with successfuly");
            }
            VMwareVirtualMachine.logger.info("VM " + this.nameLabel + " live migration done");

            this.host.postMigrateVM(this, targetVMwareHost);
            this.host = targetVMwareHost;

            this.emitNotification(NotificationTypes.VM_MIGRATION, targetHostName, this.uuid);
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("Error migrating VM", ex);
            throw new VMMException("Migration failure", ex);
        } finally {
            connection.release();
        }

    }

    public float[] getLoadPerVCPU() {
        throw new UnsupportedOperationException("not implemented");
    }

    public String getUuid() {
        if (this.uuid == null) {
            final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
            try {
                this.uuid = VMwareVirtualMachine.getUuid(connection, this.ref);
            } finally {
                connection.release();
            }
        }
        return this.uuid;
    }

    public static String getUuid(final VMwareServiceConnection connection, final ManagedObjectReference ref) {
        String uuid = "";
        try {
            DynamicProperty[] props = VMwareVirtualMachine.getDynamicProarray(connection, ref, "config.uuid");
            if (props == null || props.length < 1) {
                VMwareVirtualMachine.logger.error("Cannot retrieve UUID of VM ");
                uuid = "-1";
            } else {
                uuid = (String) props[0].getVal();
            }
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getUUID()] : ", ex);
        }
        return uuid;
    }

    public Date getStartTime() {
        VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getStartTime()] : Not Implemented ! ");
        return null;
    }

    public HostMXBean getHostMBean() {
        return this.host;
    }

    public String getNameLabel() {
        return this.nameLabel;
    }

    @Override
    public Map<String, String> getUserMetadata() {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public void createUserMetadata(final Map<String, String> metadata) {
        // TODO Auto-generated method stub
    }

    public String getMacAddress() {
        if (this.macAddress == null) {
            final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
            try {
                VirtualHardware hw = (VirtualHardware) VMwareVirtualMachine.getDynamicProarray(connection, this.ref,
                    "config.hardware")[0].getVal();
                for (VirtualDevice vd : hw.getDevice()) {
                    if (vd instanceof VirtualEthernetCard) {
                        this.macAddress = ((VirtualEthernetCard) vd).getMacAddress();
                        break;
                    }
                }
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getMacAddress()] : ", ex);
            } finally {
                connection.release();
            }
        }
        return this.macAddress;
    }

    public String getGuestIpAddress() {
        // XX do not cache IP address because it might change
        // over the VM lifetime. In particular, VMwareTools might return
        // an IP address of 0.0.0.0 upon start-up
        // if (this.ipAddress == null)
        {
            // first try to get IP address thru VMware API
            // assuming that VMware Tools are running in the guest
            int tryNumber = 72;
            final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
            try {
                while (tryNumber-- > 0) {
                    VMwareVirtualMachine.logger.debug("Retrieving IP address of VM " + this.nameLabel + " ...");
                    DynamicProperty[] props = VMwareVirtualMachine.getDynamicProarray(connection, this.ref, "guest.ipAddress");
                    if (props != null && props.length > 0) {
                        this.ipAddress = (String) props[0].getVal();
                        VMwareVirtualMachine.logger.debug("IP address of VM " + this.nameLabel + " is " + this.ipAddress);
                        break;
                    }
                    Thread.sleep(10 * 1000);
                }
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getGuestIpAddress()] : ", ex);
            } finally {
                connection.release();
            }
        }
        return this.ipAddress;
    }

    public GuestInfo getGuestInfo() {
        GuestInfo guestInfo = new GuestInfo();
        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            com.vmware.vim.GuestInfo info = (com.vmware.vim.GuestInfo) connection.getDynamicProperties(this.ref, "guest")[0]
                .getVal();
            guestInfo.setHostName(info.getHostName());
            if (info.getIpAddress() != null) {
                guestInfo.setIpAddresses(Collections.singletonList(info.getIpAddress()));
            }
            Map<String, String> properties = new HashMap<String, String>();
            guestInfo.setProperties(properties);
            if (info.getGuestFullName() != null) {
                guestInfo.setOsType(info.getGuestFullName());
            }
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getGuestInfo()] : ", ex);
        } finally {
            connection.release();
        }
        return guestInfo;
    }

    public float getCPULoad() {
        // FIXME only works once perf monitor is started
        if (this.metricValues != null && this.metricValues.get("cpu.usage") != null) {
            return (this.metricValues.get("cpu.usage").floatValue() / 10000);
        }
        return 0;
    }

    public long getUpTimeSeconds() {
        if (this.metricValues != null && this.metricValues.get("sys.uptime") != null) {
            return this.metricValues.get("sys.uptime").longValue();
        }
        return 0;
    }

    public long getMemorySizeMB() {
        if (this.memorySizeMB == -1) {
            final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
            try {
                this.memorySizeMB = (Integer) VMwareVirtualMachine.getDynamicProarray(connection, this.ref,
                    "config.hardware.memoryMB")[0].getVal();
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getMemorySizeMB()] : ", ex);
            } finally {
                connection.release();
            }
        }
        return this.memorySizeMB;
    }

    public void setMemorySizeMB(final long size) {
        VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.setMemorySizeMB(...)] : Not Implemented ! ");
    }

    public long getMemoryUsedMB() {
        if (this.metricValues != null) {
            return this.metricValues.get("mem.consumed").longValue() / 1024;
        }
        return 0;
    }

    public int getNumVCPUs() {
        if (this.numVcpus == -1) {
            final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
            try {
                this.numVcpus = (Integer) VMwareVirtualMachine.getDynamicProarray(connection, this.ref,
                    "config.hardware.numCPU")[0].getVal();
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getNumVCPUs()] : ", ex);
            } finally {
                connection.release();
            }
        }
        return this.numVcpus;
    }

    public void setNumVCPUs(final int numVCPUs) {
        // TODO
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean[][] getCPUAffinity() {
        int numVCPUs = this.getNumVCPUs();
        int numPCPUs = this.host.getNumCPU();
        boolean[][] result = new boolean[numVCPUs][];
        for (int i = 0; i < numVCPUs; i++) {
            result[i] = new boolean[numPCPUs];
        }
        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            int[] affinity = null;
            DynamicProperty[] props = VMwareVirtualMachine.getDynamicProarray(connection, this.ref,
                "config.cpuAffinity.affinitySet");
            if (props != null && props.length > 0) {
                affinity = ((ArrayOfInt) props[0].getVal()).get_int();
            }
            if (affinity == null || affinity.length == 0) {
                for (int i = 0; i < numVCPUs; i++) {
                    for (int j = 0; j < numPCPUs; j++) {
                        result[i][j] = true;
                    }
                }
            } else {
                for (int i = 0; i < numVCPUs; i++) {
                    for (int j = 0; j < affinity.length; j++) {
                        result[i][affinity[j]] = true;
                    }
                }
            }
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getCPUAffinity()] : ", ex);
        } finally {
            connection.release();
        }
        return result;
    }

    @Override
    public void setCPUAffinity(final boolean[][] affinity) {
        // TODO
    }

    public int getSchedulingCap() {
        // TODO
        int result = 0;
        VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getSchedulingCap()] : Not Implemented ! ");
        return result;
    }

    public void setSchedulingCap(final int schedulingCap) {
        // TODO
        VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.setSchedulingCap(...)] : Not Implemented ! ");
    }

    public int getSchedulingWeight() {
        // TODO
        int result = 0;
        VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getSchedulingWeight()] : Not Implemented ! ");
        return result;
    }

    public void setSchedulingWeight(final int schedulingWeight) {
        // TODO
        VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.setSchedulingWeight(...)] : Not Implemented ! ");
    }

    public PowerState getState() {
        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            DynamicProperty props[] = VMwareVirtualMachine.getDynamicProarray(connection, this.ref, "runtime.powerState");
            if (props != null) {
                VirtualMachinePowerState state = (VirtualMachinePowerState) props[0].getVal();
                if (state.getValue().equals("poweredOff")) {
                    return PowerState.HALTED;
                } else if (state.getValue().equals("poweredOn")) {
                    return PowerState.RUNNING;
                } else if (state.getValue().equals("suspended")) {
                    return PowerState.SUSPENDED;
                }
            } else {
                return PowerState.UNKNOWN;
            }
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("[VMWareVirtualMachine.getState()]: VM " + this.getNameLabel(), ex);
        } finally {
            connection.release();
        }

        return PowerState.UNKNOWN;
    }

    private PowerState lastState = PowerState.UNKNOWN;

    void onPowerStateChangedVM(final VirtualMachineMXBean.PowerState state, final long time) {
        if (state != this.lastState) {
            this.emitNotification(NotificationTypes.VM_STATE_CHANGE, state.toString(), null);
            VMwareVirtualMachine.logger.debug("VM " + this.getNameLabel() + " new power state: " + state.toString());
            this.lastState = state;
        }
    }

    public String getConsole() throws VMMException {
        if (this.host.hostConnectionPool == null) {
            return null;
        }
        final VMwareServiceConnection connection = this.host.hostConnectionPool.getConnection();
        try {
            ManagedObjectReference searchIndexService = connection.getServiceContent().getSearchIndex();
            ManagedObjectReference vmRefOnHost = connection.getService().findByUuid(searchIndexService, null, this.getUuid(),
                true, false);
            if (vmRefOnHost == null) {
                VMwareVirtualMachine.logger.error("Cannot findByUuid VM " + this.nameLabel + " uuid=" + this.uuid);
                throw new VMMException("Cannot retrieve VM object id on host");
            }
            ManagedObjectReference sessionManager = connection.getServiceContent().getSessionManager();
            String ticket = connection.getService().acquireCloneTicket(sessionManager);
            VMwareVirtualMachine.logger.debug("vmid=" + vmRefOnHost.get_value() + " clone ticket=" + ticket);
            return "vmware-vmrc://" + this.host.getHostName() + "?id=" + vmRefOnHost.get_value() + ",ticket=" + ticket;
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("VM " + this.nameLabel + " getConsole failed ", ex);
            VMwareHelper.translateException("Failed to acquire clone ticket for VM console", ex);
            return null;
        } finally {
            connection.release();
        }
    }

    public void addUserData(final String key, final String value) {
        throw new UnsupportedOperationException("not yet implemented");
    }

    public void suspend() {
        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            ManagedObjectReference taskmor = connection.getService().suspendVM_Task(this.ref);
            boolean result = this.getTaskInfo(connection, taskmor);
            if (result) {
                VMwareVirtualMachine.logger.debug("Virtual Machine " + this.nameLabel + " suspended on successfully");
            }
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("Error suspending VM", ex);
        } finally {
            connection.release();
        }
    }

    public void resume() {
        this.start();
    }

    public void shutdown() {
        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            ManagedObjectReference taskmor = connection.getService().powerOffVM_Task(this.ref);
            boolean result = this.getTaskInfo(connection, taskmor);
            if (result) {
                VMwareVirtualMachine.logger.debug("Virtual Machine " + this.nameLabel + " powered off successfuly");
            }
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("Error shutting down VM", ex);
        } finally {
            connection.release();
        }
    }

    public void start() {
        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            ManagedObjectReference taskmor = connection.getService().powerOnVM_Task(this.ref, null);
            boolean result = this.getTaskInfo(connection, taskmor);
            if (result) {
                VMwareVirtualMachine.logger.debug("Virtual Machine " + this.nameLabel + " powered on successfully");
            }
        } catch (InvalidPowerState e) {
            VMwareVirtualMachine.logger.debug("Virtual Machine is already powered on");
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("Error starting VM", ex);
        } finally {
            connection.release();
        }
    }

    public void pause() {
        throw new UnsupportedOperationException();
    }

    public void unpause() {
        throw new UnsupportedOperationException();
    }

    public void reboot() {
        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            // TODO this is a hard reboot
            // should we use reboot_guest ?
            connection.getService().resetVM_Task(this.ref);
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("Error rebooting VM", ex);
        } finally {
            connection.release();
        }
    }

    public void destroy() {
        this.destroy(true);
    }

    public void destroy(final boolean deleteDisks) {
        if (this.getState() == PowerState.RUNNING) {
            this.shutdown();
        }
        // TODO remove disks from the VM
        if (!deleteDisks) {

        }

        final VMwareServiceConnection connection = this.pool.getConnectionPool().getConnection();
        try {
            ManagedObjectReference taskmor = connection.getService().destroy_Task(this.ref);
            boolean result = this.getTaskInfo(connection, taskmor);
            if (result) {
                VMwareVirtualMachine.logger.debug("Virtual Machine " + this.nameLabel + " destroyed successfully");
            }

        } catch (InvalidPowerState e) {
            VMwareVirtualMachine.logger.debug("Invalid power state");
        } catch (Exception ex) {
            VMwareVirtualMachine.logger.error("Error destroying VM", ex);
        } finally {
            connection.release();
        }
    }

    private static DynamicProperty[] getDynamicProarray(final VMwareServiceConnection connection,
        final ManagedObjectReference MOR, final String pName) throws Exception {
        ObjectContent[] objContent;
        objContent = connection.getObjectProperties(null, MOR, new String[] {pName});
        if (objContent != null) {
            ObjectContent contentObj = objContent[0];
            DynamicProperty[] objArr = contentObj.getPropSet();
            return objArr;
        } else {
            return null;
        }
    }

    private boolean getTaskInfo(final VMwareServiceConnection connection, final ManagedObjectReference taskmor)
        throws Exception {
        boolean valid = false;
        String res;
        res = connection.waitForTask(taskmor);
        if (res.equalsIgnoreCase("success")) {
            valid = true;
        } else {
            valid = false;
        }
        return valid;
    }

    // TODO
    public VirtualMachineMXBean cloneVM(final String name, final VMCustomizationSpec custSpec, final boolean sync)
        throws InsufficientResourcesException {
        throw new UnsupportedOperationException("Not implemented");
    }

    // TODO store name & description in user data fields of the VMware template
    @Override
    public VirtualMachineImageMXBean makeTemplate(final String name, final String description,
        final Map<String, String> metadata) throws InsufficientResourcesException, IllegalOperationException,
        BadVMPowerStateException, VMMException {

        String vmImageID = name + "-" + System.currentTimeMillis();

        VMwareServiceConnection connection = this.host.getConnectionPool().getConnection();
        if (connection == null) {
            VMwareVirtualMachine.logger.debug("Out of connection");
            throw new VMMException("Too many VM creation requests, try again later");
        }

        VirtualMachineCloneSpec cloneSpec = new VirtualMachineCloneSpec();
        VirtualMachineRelocateSpec relocSpec = new VirtualMachineRelocateSpec();
        try {

            this.shutdown();

            relocSpec.setPool(this.host.getVCResourcePool(connection));
            relocSpec.setDatastore(this.pool.getDatastoreRef());
            cloneSpec.setLocation(relocSpec);
            cloneSpec.setPowerOn(false);
            cloneSpec.setTemplate(true);

            synchronized (connection) {
                VMwareVirtualMachine.logger.debug("Making template " + vmImageID + " from VM " + this.getNameLabel());
                String status = "";
                try {
                    ManagedObjectReference cloneTask = connection.getService().cloneVM_Task(this.ref,
                        this.pool.getVmTemplateFolderRef(), vmImageID, cloneSpec);
                    status = connection.waitForTask(cloneTask);
                } catch (Exception e) {
                    VMwareVirtualMachine.logger.error("Failed to create template", e);
                    throw new VMMException(e);
                }

                VMwareVirtualMachine.logger.debug("CloneVMTask, status=" + status);

                if (status.equalsIgnoreCase("failure")) {
                    String msg = "Failure -: Virtual Machine cannot be cloned to template";
                    VMwareVirtualMachine.logger.error(msg);
                    throw new VMMException(msg);
                }
            }
        } finally {
            connection.release();
            this.start();
        }

        ((VMwareServerPool.VMwareVMTemplateDataStore) this.host.getVMImageStore()).sync();
        return ((VMwareServerPool.VMwareVMTemplateDataStore) this.host.getVMImageStore()).lookUpByUUID(vmImageID);

    }

    @Override
    public Volume makeImage(final String name, final StoragePoolMXBean storagePool) throws VMMException {
        VMwareStoragePool vmwareStoragePool = (VMwareStoragePool) this.pool.getStoragePoolByName(storagePool.getName());

        String templateName = name + "-" + System.nanoTime();

        VMwareServiceConnection connection = this.host.getConnectionPool().getConnection();

        try {
            VirtualMachineCloneSpec cloneSpec = new VirtualMachineCloneSpec();
            VirtualMachineRelocateSpec relocSpec = new VirtualMachineRelocateSpec();

            relocSpec.setPool(this.host.getVCResourcePool(connection));
            relocSpec.setDatastore(vmwareStoragePool.getDataStore());
            cloneSpec.setLocation(relocSpec);
            cloneSpec.setPowerOn(false);
            cloneSpec.setTemplate(true);

            VMwareVirtualMachine.logger.debug("Making template " + templateName + " from VM " + this.getNameLabel());
            ManagedObjectReference task;
            String status;
            TaskInfo taskInfo = null;
            try {
                task = connection.getService().cloneVM_Task(this.ref, this.pool.getVmTemplateFolderRef(), templateName,
                    cloneSpec);
                status = connection.waitForTask(task);
                taskInfo = (TaskInfo) connection.getDynamicProperties(task, "info")[0].getVal();
            } catch (Exception e) {
                VMwareVirtualMachine.logger.error("Failed to create template", e);
                throw new VMMException(e);
            }
            if (!status.equalsIgnoreCase("success")) {
                throw new VMMException("Failed to create image:" + taskInfo.getError().getLocalizedMessage());
            }
            ManagedObjectReference templateVm = (ManagedObjectReference) taskInfo.getResult();
            try {
                List<Volume> vols = vmwareStoragePool.addVMVolumes(connection, templateVm, true);
                return vols.get(0);
            } catch (Exception e) {
                VMwareVirtualMachine.logger.error("Failed to create template", e);
                throw new VMMException(e);
            }
        } finally {
            connection.release();
        }
    }

    @Override
    public List<VirtualDisk> getVirtualDisks() {
        if (this.disks == null) {
            this.disks = new ArrayList<VirtualDisk>();
            List<StoragePoolMXBean> storagePools = ((VMwareServerPool) this.host.getServerPool()).getStoragePools();
            VMwareServiceConnection connection = this.host.getConnectionPool().getConnection();
            try {
                ArrayOfVirtualDevice devices = (ArrayOfVirtualDevice) connection.getDynamicProperties(this.ref,
                    "config.hardware.device")[0].getVal();
                for (VirtualDevice device : devices.getVirtualDevice()) {
                    if (device instanceof com.vmware.vim.VirtualDisk) {
                        if (device.getBacking() instanceof VirtualDeviceFileBackingInfo) {
                            VirtualDeviceFileBackingInfo fileBacking = (VirtualDeviceFileBackingInfo) device.getBacking();
                            // XXX fileBacking.fileName includes datastore
                            // prefix ?
                            // XXX ignore disks not in storagepool

                            ManagedObjectReference diskDataStore = fileBacking.getDatastore();
                            for (StoragePoolMXBean pool : storagePools) {
                                VMwareStoragePool vmwarePool = (VMwareStoragePool) pool;
                                if (diskDataStore.equals(vmwarePool.getDataStore())) {
                                    Volume vol = vmwarePool.getVolumeByKey(fileBacking.getFileName());
                                    if (vol != null) {
                                        VirtualDisk disk = new VirtualDisk();
                                        disk.setVolume(vol);
                                        disk.setDevice("");
                                        this.disks.add(disk);
                                    }
                                    break;
                                }
                            }
                        }
                    }
                }
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("Failed to retrieve disks of VM " + this.nameLabel, ex);
            } finally {
                connection.release();
            }
        }
        return this.disks;
    }

    @Override
    public List<org.ow2.sirocco.vmm.api.VirtualCdrom> getVirtualCdroms() {
        if (this.cdroms == null) {
            this.cdroms = new ArrayList<org.ow2.sirocco.vmm.api.VirtualCdrom>();
            List<StoragePoolMXBean> storagePools = ((VMwareServerPool) this.host.getServerPool()).getStoragePools();
            VMwareServiceConnection connection = this.host.getConnectionPool().getConnection();
            try {
                ArrayOfVirtualDevice devices = (ArrayOfVirtualDevice) connection.getDynamicProperties(this.ref,
                    "config.hardware.device")[0].getVal();
                VirtualCdrom cdromDevice = null;
                int deviceId = 0;
                for (VirtualDevice device : devices.getVirtualDevice()) {
                    if (device instanceof VirtualCdrom) {
                        cdromDevice = (VirtualCdrom) device;
                        org.ow2.sirocco.vmm.api.VirtualCdrom cdrom = new org.ow2.sirocco.vmm.api.VirtualCdrom();
                        cdrom.setDevice("" + deviceId++);
                        if (cdromDevice.getBacking() != null
                            && cdromDevice.getBacking() instanceof VirtualDeviceFileBackingInfo) {
                            VirtualDeviceFileBackingInfo fileBacking = (VirtualDeviceFileBackingInfo) cdromDevice.getBacking();
                            ManagedObjectReference diskDataStore = fileBacking.getDatastore();
                            if (cdromDevice.getConnectable().isConnected()) {
                                for (StoragePoolMXBean pool : storagePools) {
                                    VMwareStoragePool vmwarePool = (VMwareStoragePool) pool;
                                    if (diskDataStore.equals(vmwarePool.getDataStore())) {
                                        Volume vol = vmwarePool.getVolumeByKey(fileBacking.getFileName());
                                        if (vol != null) {
                                            cdrom.setVolume(vol);
                                            break;
                                        }
                                    }
                                }
                            }
                            this.cdroms.add(cdrom);
                        }
                    }
                }
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("Failed to get CD-ROM devices of VM " + this.nameLabel, ex);
            } finally {
                connection.release();
            }
        }
        return this.cdroms;
    }

    @Override
    public void attachVolume(final VirtualDiskSpec diskSpec) throws VMMException {
        if (diskSpec.getDiskOp() != DiskOperation.ATTACH) {
            throw new VMMException("Operation not supported");
        }
        // VMwareStoragePool storagePool = (VMwareStoragePool)
        // this.pool.getStoragePoolByName(diskSpec.getStoragePool().getName());

        VirtualSCSIController scsiController = null;
        int unitNumber;
        VMwareServiceConnection connection = this.host.getConnectionPool().getConnection();
        try {
            try {
                ArrayOfVirtualDevice devices = (ArrayOfVirtualDevice) connection.getDynamicProperties(this.ref,
                    "config.hardware.device")[0].getVal();
                unitNumber = devices.getVirtualDevice().length + 1;
                for (VirtualDevice device : devices.getVirtualDevice()) {
                    if (device instanceof VirtualSCSIController && scsiController == null) {
                        scsiController = (VirtualSCSIController) device;
                        break;
                    }
                }
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("Failed to get devices", ex);
                throw new VMMException("Failed to attach volume " + ex.getMessage());
            }
            VirtualDeviceConfigSpec deviceSpec = new VirtualDeviceConfigSpec();
            deviceSpec.setOperation(VirtualDeviceConfigSpecOperation.add);
            com.vmware.vim.VirtualDisk disk = new com.vmware.vim.VirtualDisk();
            VirtualDiskFlatVer2BackingInfo diskfileBacking = new VirtualDiskFlatVer2BackingInfo();
            diskfileBacking.setFileName(diskSpec.getVolume().getPath());
            diskfileBacking.setDiskMode("persistent");
            disk.setControllerKey(scsiController.getKey());
            disk.setUnitNumber(unitNumber++);
            disk.setKey(-1);
            disk.setBacking(diskfileBacking);
            disk.setCapacityInKB(diskSpec.getCapacityMB() * 1024);
            deviceSpec.setDevice(disk);

            VirtualMachineConfigSpec vmReConfigSpec = new VirtualMachineConfigSpec();
            vmReConfigSpec.setDeviceChange(new VirtualDeviceConfigSpec[] {deviceSpec});

            String status;
            TaskInfo taskInfo;
            try {
                ManagedObjectReference task = connection.getService().reconfigVM_Task(this.ref, vmReConfigSpec);
                status = connection.waitForTask(task);
                taskInfo = (TaskInfo) connection.getDynamicProperties(task, "info")[0].getVal();
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("Failed reconfigure VM ", ex);
                throw new VMMException("Failed to change CD-ROM media: " + ex.getMessage());
            }
            if (!status.equalsIgnoreCase("success")) {
                throw new VMMException("Failed to reconfigure VM " + taskInfo.getError().getLocalizedMessage());
            }
            VirtualDisk newDisk = new VirtualDisk();
            newDisk.setDevice("");
            newDisk.setVolume(diskSpec.getVolume());
            this.disks.add(newDisk);

        } finally {
            connection.release();
        }
    }

    @Override
    public void detachVolume(final String key) throws VMMException {
        VirtualDisk diskToDetach = null;
        for (VirtualDisk disk : this.getVirtualDisks()) {
            if (disk.getVolume().getKey().equals(key)) {
                this.disks.remove(disk);
                diskToDetach = disk;
                break;
            }
        }
        if (diskToDetach == null) {
            throw new VMMException("Invalid volume key " + key);
        }

        VMwareServiceConnection connection = this.host.getConnectionPool().getConnection();
        try {
            com.vmware.vim.VirtualDisk deviceToRemove = null;
            try {
                ArrayOfVirtualDevice devices = (ArrayOfVirtualDevice) connection.getDynamicProperties(this.ref,
                    "config.hardware.device")[0].getVal();
                for (VirtualDevice device : devices.getVirtualDevice()) {
                    if (device instanceof com.vmware.vim.VirtualDisk) {
                        com.vmware.vim.VirtualDisk disk = (com.vmware.vim.VirtualDisk) device;
                        if (disk.getBacking() != null && disk.getBacking() instanceof VirtualDeviceFileBackingInfo) {
                            VirtualDeviceFileBackingInfo fileBacking = (VirtualDeviceFileBackingInfo) disk.getBacking();
                            if (fileBacking.getFileName().equals(key)) {
                                deviceToRemove = disk;
                                break;
                            }
                        }
                    }
                }
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("Failed to get devices", ex);
                throw new VMMException("Failed to attach volume " + ex.getMessage());
            }
            if (deviceToRemove == null) {
                throw new VMMException("Invalid volume key " + key);
            }

            VirtualDeviceConfigSpec deviceSpec = new VirtualDeviceConfigSpec();
            deviceSpec.setOperation(VirtualDeviceConfigSpecOperation.remove);
            deviceSpec.setDevice(deviceToRemove);

            VirtualMachineConfigSpec vmReConfigSpec = new VirtualMachineConfigSpec();
            vmReConfigSpec.setDeviceChange(new VirtualDeviceConfigSpec[] {deviceSpec});

            String status;
            TaskInfo taskInfo;
            try {
                ManagedObjectReference task = connection.getService().reconfigVM_Task(this.ref, vmReConfigSpec);
                status = connection.waitForTask(task);
                taskInfo = (TaskInfo) connection.getDynamicProperties(task, "info")[0].getVal();
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("Failed to reconfigure VM ", ex);
                throw new VMMException("Failed to detach volume: " + ex.getMessage());
            }
            if (!status.equalsIgnoreCase("success")) {
                throw new VMMException("Failed to detach volume " + taskInfo.getError().getLocalizedMessage());
            }

        } finally {
            connection.release();
        }
    }

    @Override
    public void changeCdromMedia(final String deviceId, final Volume iso) throws VMMException {
        VMwareServiceConnection connection = this.host.getConnectionPool().getConnection();
        try {
            VirtualCdrom cdromDevice = null;
            try {
                ArrayOfVirtualDevice devices = (ArrayOfVirtualDevice) connection.getDynamicProperties(this.ref,
                    "config.hardware.device")[0].getVal();

                int cdromIndex = 0;
                for (VirtualDevice device : devices.getVirtualDevice()) {
                    if (device instanceof VirtualCdrom) {
                        if (cdromIndex++ == Integer.valueOf(deviceId)) {
                            cdromDevice = (VirtualCdrom) device;
                            break;
                        }
                    }
                }
            } catch (Exception ex) {
                VMwareVirtualMachine.logger.error("Failed get devices ", ex);
                throw new VMMException("Failed to change CD-ROM media: " + ex.getMessage());
            }
            if (cdromDevice != null) {
                VirtualDeviceConfigSpec cdromDeviceSpec = new VirtualDeviceConfigSpec();
                cdromDeviceSpec.setOperation(VirtualDeviceConfigSpecOperation.edit);
                VirtualDeviceConnectInfo connectInfo = new VirtualDeviceConnectInfo();

                if (iso != null) {
                    VMwareStoragePool storagePool = (VMwareStoragePool) this.pool.getStoragePoolByName(iso.getStoragePool()
                        .getName());
                    if (cdromDevice.getBacking() != null && cdromDevice.getBacking() instanceof VirtualCdromIsoBackingInfo) {
                        VirtualCdromIsoBackingInfo cdromBacking = (VirtualCdromIsoBackingInfo) cdromDevice.getBacking();
                        cdromBacking.setFileName(iso.getPath());
                        cdromBacking.setDatastore(storagePool.getDataStore());
                    }
                    // cdromDeviceSpec.setFileOperation(VirtualDeviceConfigSpecFileOperation.replace);
                    // VirtualCdromIsoBackingInfo cdromBacking = new
                    // VirtualCdromIsoBackingInfo();
                    // cdromBacking.setDatastore(storagePool.getDataStore());
                    // cdromBacking.setFileName(iso.getPath());
                    // cdromDevice.setBacking(cdromBacking);
                    connectInfo.setConnected(true);
                } else {
                    connectInfo.setConnected(false);
                }
                cdromDevice.setConnectable(connectInfo);
                cdromDeviceSpec.setDevice(cdromDevice);
                VirtualMachineConfigSpec vmReConfigSpec = new VirtualMachineConfigSpec();
                vmReConfigSpec.setDeviceChange(new VirtualDeviceConfigSpec[] {cdromDeviceSpec});

                String status;
                TaskInfo taskInfo;
                try {
                    ManagedObjectReference task = connection.getService().reconfigVM_Task(this.ref, vmReConfigSpec);
                    status = connection.waitForTask(task);
                    taskInfo = (TaskInfo) connection.getDynamicProperties(task, "info")[0].getVal();
                } catch (Exception ex) {
                    VMwareVirtualMachine.logger.error("Failed get devices ", ex);
                    throw new VMMException("Failed to change CD-ROM media: " + ex.getMessage());
                }
                if (!status.equalsIgnoreCase("success")) {
                    throw new VMMException("Failed to reconfigure VM " + taskInfo.getError().getLocalizedMessage());
                }
                for (org.ow2.sirocco.vmm.api.VirtualCdrom cdrom : this.getVirtualCdroms()) {
                    if (cdrom.getDevice().equals(deviceId)) {
                        cdrom.setVolume(iso);
                        break;
                    }
                }
            } else {
                throw new VMMException("Cannot find CD-ROM device " + deviceId);
            }

        } finally {
            connection.release();
        }
    }

    @Override
    public List<VNIC> getVNICs() {
        // TODO Auto-generated method stub
        return null;
    }
}
